@@ -1,5 +1,5 @@
 #!/usr/bin/env python3
-#-*- coding: utf-8 -*-
+# -*- coding: utf-8 -*-
 # pylint: disable=invalid-name, protected-access
 
 """
@@ -28,28 +28,96 @@ from configparser import ConfigParser
 
 import click
 import psycopg2
+from git import Repo, GitCommandError
 
 from half_orm.model import Model, CONF_DIR
 from half_orm.model_errors import MissingConfigFile
 
-from half_orm_packager.globals import TEMPLATES_DIR, get_connection_file_name, hop_version
+from half_orm_packager.globals import TEMPLATES_DIR, hop_version
 from half_orm_packager.patch import Patch
 from half_orm_packager.test import tests
+from half_orm_packager.update import update_modules
+from half_orm_packager.hgit import HGit
+
+
+def get_connection_file_name(base_dir=None, ref_dir=None):
+    """searches the hop configuration file for the package.
+    This method is called when no hop config file is provided.
+    It changes to the package base directory if the config file exists.
+    """
+    config = ConfigParser()
+
+    cur_dir = base_dir
+    if not base_dir:
+        ref_dir = os.path.abspath(os.path.curdir)
+        cur_dir = base_dir = ref_dir
+    for base in ['hop', 'halfORM']:
+        if os.path.exists('.{}/config'.format(base)):
+            config.read('.{}/config'.format(base))
+            config_file = config['halfORM']['config_file']
+            package_name = config['halfORM']['package_name']
+            return config_file, package_name, cur_dir
+
+    if os.path.abspath(os.path.curdir) != '/':
+        os.chdir('..')
+        cur_dir = os.path.abspath(os.path.curdir)
+        return get_connection_file_name(cur_dir, ref_dir)
+    # restore reference directory.
+    os.chdir(ref_dir)
+    return None, None, None
 
 class Hop:
-    connection_file_name, package_name = get_connection_file_name()
+    "XXX: The hop class doc..."
+    __connection_file_name, __package_name, __project_path = get_connection_file_name()
+    __model = None
 
-    def alpha(self, model):
+    @property
+    def connection_file_name(self):
+        "returns the connection file name"
+        return self.__connection_file_name
+
+    @property
+    def package_name(self):
+        "returns the package name"
+        return self.__package_name
+
+    @package_name.setter
+    def package_name(self, package_name):
+        self.__package_name = package_name
+
+    @property
+    def project_path(self):
+        return self.__project_path
+
+    @project_path.setter
+    def project_path(self, project_path):
+        if self.__project_path is None:
+            self.__project_path = project_path
+
+    @property
+    def package_path(self):
+        return f'{self.project_path}/{self.package_name}'
+
+    @property
+    def model(self):
+        "model getter"
+        if self.__model is None:
+            self.model = get_model()
+        return self.__model
+
+    @model.setter
+    def model(self, model):
+        "model setter"
+        self.__model = model
+
+    def alpha(self):
         """Toutes les modifs à faire durant la mise au point de hop
         """
-        if not model.has_relation('half_orm_meta.hop_release'):
-            try:
-                model.get_relation_class('meta.release')
-                model.get_relation_class('meta.release_issue')
-                model.get_relation_class('meta.view.last_release')
-                model.get_relation_class('meta.view.penultimate_release')
-                click.echo("ALPHA: Renaming meta.release to half_orm_meta.hop_release, ...")
-                model.execute_query("""
+        if not self.model.has_relation('half_orm_meta.hop_release'):
+            if self.model.has_relation('meta.release'):
+                click.echo(
+                    "ALPHA: Renaming meta.release to half_orm_meta.hop_release, ...")
+                self.model.execute_query("""
                 create schema half_orm_meta;
                 create schema "half_orm_meta.view";
                 alter table meta.release set schema half_orm_meta;
@@ -63,20 +131,17 @@ class Hop:
                 """)
                 click.echo("Please re-run the command.")
                 sys.exit()
-            except Exception as err:
-                print('ALPHA ERR', err)
-                pass
         # if not model.has_relation('half_orm_meta.view.hop_penultimate_release'):
         #     TODO: fix missing penultimate_release on some databases.
         return Model(HOP.package_name)
 
-
     def __str__(self):
         return f"""
         connection_file_name: {self.connection_file_name}
         package_name: {self.package_name}
         """
 
+
 HOP = Hop()
 # print(HOP)
 
@@ -91,80 +156,68 @@ port = {port}
 production = {production}
 """
 
+
 def status():
+    """Prints the status"""
     print('STATUS')
-    model = get_model()
-    next_release = Patch(model).get_next_release()
+    print(HOP)
+    next_release = Patch(HOP).get_next_release()
     while next_release:
-        next_release = Patch(model).get_next_release(next_release)
+        next_release = Patch(HOP).get_next_release(next_release)
     print('hop --help to get help.')
 
+
+def read_template(file_path):
+    "helper"
+    with open(file_path, encoding='utf-8') as file_:
+        return file_.read()
+
+def write_file(file_path, content):
+    "helper"
+    with open(file_path, 'w', encoding='utf-8') as file_:
+        file_.write(content)
+
 def init_package(model, project_name: str):
     """Initialises the package directory.
 
     model (Model): The loaded model instance
     project_name (str): The project name (hop create argument)
     """
-    from git import Repo, GitCommandError
-
     curdir = os.path.abspath(os.curdir)
-    os.chdir(TEMPLATES_DIR)
-    README = open('README').read()
-    CONFIG_TEMPLATE = open('config').read()
-    SETUP_TEMPLATE = open('setup.py').read()
-    GIT_IGNORE = open('.gitignore').read()
-    PIPFILE = open('Pipfile').read()
     project_path = os.path.join(curdir, project_name)
     if not os.path.exists(project_path):
         os.makedirs(project_path)
     else:
-        raise Exception(f'The path {project_path} already exists')
-
-    os.chdir(project_path)
+        sys.stderr.write(f"The path '{project_path}' already exists!\n")
+        sys.exit(1)
+    README = read_template(f'{TEMPLATES_DIR}/README')
+    CONFIG_TEMPLATE = read_template(f'{TEMPLATES_DIR}/config')
+    SETUP_TEMPLATE = read_template(f'{TEMPLATES_DIR}/setup.py')
+    GIT_IGNORE = read_template(f'{TEMPLATES_DIR}/.gitignore')
+    PIPFILE = read_template(f'{TEMPLATES_DIR}/Pipfile')
 
     dbname = model._dbname
     setup = SETUP_TEMPLATE.format(dbname=dbname, package_name=project_name)
-    open('./setup.py', 'w').write(setup)
-    open('./Pipfile', 'w').write(PIPFILE)
-    os.makedirs('./.hop')
-    open(f'./.hop/config', 'w').write(
+    write_file(f'{project_path}/setup.py', setup)
+    write_file(f'{project_path}/Pipfile', PIPFILE)
+    os.mkdir(f'{project_path}/.hop')
+    write_file(f'{project_path}/.hop/config',
         CONFIG_TEMPLATE.format(
             config_file=project_name, package_name=project_name))
     cmd = " ".join(sys.argv)
     readme = README.format(cmd=cmd, dbname=dbname, package_name=project_name)
-    open('./README.md', 'w').write(readme)
-    open('./.gitignore', 'w').write(GIT_IGNORE)
-    os.mkdir(f'./{project_name}')
-    try:
-        Repo.init('.', initial_branch='main')
-        print("Initializing git with a 'main' branch.")
-    except GitCommandError:
-        Repo.init('.')
-        print("Initializing git with a 'master' branch.")
-
-    repo = Repo('.')
-    Patch(model, create_mode=True).patch()
-    model.reconnect() # we get the new stuff from db metadata here
-    subprocess.run(['hop', 'update', '-f']) # hop creates/updates the modules & ignore tests
-
-    try:
-        repo.head.commit
-    except ValueError:
-        repo.git.add('.')
-        repo.git.commit(m='[0.0.0] First release')
-
-    print("Switching to the 'devel' branch.")
-    repo.git.checkout(b='devel')
-
-@property
-def package_name(self):
-    return self.__config.get('package_name')
-
+    write_file(f'{project_path}/README.md', readme)
+    write_file(f'{project_path}/.gitignore', GIT_IGNORE)
+    os.mkdir(f'{project_path}/{project_name}')
+    HOP.project_path = project_path
+    HGit(HOP).init()
+    print(f"\nThe hop project '{project_name}' has been created.")
 
 def set_config_file(project_name: str):
     """ Asks for the connection parameters. Returns a dictionary with the params.
     """
-
+    print(f'HALFORM_CONF_DIR: {CONF_DIR}')
+    HOP.package_name = project_name
     conf_path = os.path.join(CONF_DIR, project_name)
     if not os.path.isfile(conf_path):
         if not os.access(CONF_DIR, os.W_OK):
@@ -174,17 +227,18 @@ def set_config_file(project_name: str):
                     "Set the HALFORM_CONF_DIR environment variable if you want to use a\n"
                     "different directory.\n")
             sys.exit(1)
-        dbname = input(f'Database ({project_name}): ') or project_name
-        print(f'Input the connection parameters to the {dbname} database.')
+        print('Connection parameters to the database:')
+        dbname = input(f'. database name ({project_name}): ') or project_name
         user = os.environ['USER']
-        user = input(f'User ({user}): ') or user
-        password = getpass('Password: ')
+        user = input(f'. user ({user}): ') or user
+        password = getpass('. password: ')
         if password == '' and \
-            (input('Is it an ident login with a local account? [Y/n] ') or 'Y').upper() == 'Y':
-                host = port = ''
+                (input(
+                    '. is it an ident login with a local account? [Y/n] ') or 'Y').upper() == 'Y':
+            host = port = ''
         else:
-            host = input('Host (localhost): ') or 'localhost'
-            port = input('Port (5432): ') or 5432
+            host = input('. host (localhost): ') or 'localhost'
+            port = input('. port (5432): ') or 5432
 
         production = input('Production (False): ') or False
 
@@ -196,24 +250,27 @@ def set_config_file(project_name: str):
             'port': port,
             'production': production
         }
-        open(f'{CONF_DIR}/{project_name}', 'w').write(TMPL_CONF_FILE.format(**res))
-
+        open(f'{CONF_DIR}/{project_name}',
+             'w', encoding='utf-8').write(TMPL_CONF_FILE.format(**res))
+    else:
+        print(f"Using '{CONF_DIR}/{project_name}' file for connexion.")
 
     try:
         return Model(project_name)
     except psycopg2.OperationalError:
         config = ConfigParser()
-        config.read([ conf_path ])
+        config.read([conf_path])
         dbname = config.get('database', 'name')
 
-        sys.stderr.write(f'The {dbname} database does not exist.\n')
+        sys.stderr.write(f"The database '{dbname}' does not exist.\n")
         create = input('Do you want to create it (Y/n): ') or "y"
         if create.upper() == 'Y':
-            subprocess.run(['createdb', dbname])
-            model = Model(project_name)
-            return model
+            subprocess.run(['createdb', dbname], check=True)
+            return Model(project_name)
+        print(f'Please create the database an rerun hop new {project_name}')
         sys.exit(1)
 
+
 @click.group(invoke_without_command=True)
 @click.pass_context
 @click.option('-v', '--version', is_flag=True)
@@ -221,14 +278,9 @@ def main(ctx, version):
     """
     Generates/Synchronises/Patches a python package from a PostgreSQL database
     """
-    if ctx.invoked_subcommand != 'new':
-        get_model()
 
     if ctx.invoked_subcommand is None:
-        try:
-            status()
-        except:
-            pass
+        status()
     if version:
         click.echo(f'hop {hop_version()}')
         sys.exit()
@@ -256,49 +308,50 @@ def new(package_name):
 
 
 def get_model():
-
+    "Returns the half_orm model"
     # config_file, package_name = get_connection_file_name()
 
     if not HOP.package_name:
         sys.stderr.write(
-            "You're not in a halfORM package directory.\n"
+            "You're not in a hop package directory.\n"
             "Try hop --help.\n")
         sys.exit(1)
 
     try:
-        model = Model(HOP.package_name)
-        model = HOP.alpha(model) #XXX To remove after alpha
+        HOP.model = Model(HOP.package_name)
+        model = HOP.alpha()  # XXX To remove after alpha
         return model
     except psycopg2.OperationalError as exc:
         sys.stderr.write(f'The database {HOP.package_name} does not exist.\n')
         raise exc
     except MissingConfigFile:
-        sys.stderr.write(f'Cannot find the half_orm config file for this database.\n')
+        sys.stderr.write(
+            'Cannot find the half_orm config file for this database.\n')
         sys.exit(1)
 
 
-# @main.command()
-def init():
-    """ Initialize a cloned hop project by applying the base patch
-    """
-    try:
-        model = get_model()
-    except psycopg2.OperationalError as exc:
-        # config_file, package_name = get_connection_file_name()
-        model = set_config_file(HOP.package_name)
+# # @main.command()
+# def init():
+#     """ Initialize a cloned hop project by applying the base patch
+#     """
+#     try:
+#         model = get_model()
+#     except psycopg2.OperationalError:
+#         # config_file, package_name = get_connection_file_name()
+#         model = set_config_file(HOP.package_name)
 
-    Patch(model, init_mode=True).patch()
-    sys.exit()
+#     Patch(HOP, init_mode=True).patch(HOP.package_name)
+#     sys.exit()
 
 
 @main.command()
-def patch():
+@click.option('-f', '--force', is_flag=True, help="Don't check if git repo is clean.")
+@click.option('-r', '--revert', is_flag=True, help="Revert to the previous release.")
+def patch(force, revert):
     """ Applies the next patch.
     """
 
-    model = get_model()
-    Patch(model).patch()
-
+    Patch(HOP).patch(force, revert)
 
     sys.exit()
 
@@ -308,10 +361,8 @@ def patch():
 def update(force):
     """Updates the Python code with the changes made to the model.
     """
-    from half_orm_packager.update import update_modules
-    model = get_model()
-    if force or tests(model, Hop.package_name):
-        update_modules(model, Hop.package_name)
+    if force or tests(HOP.model, HOP.package_name):
+        update_modules(HOP.model, HOP.package_name)
     else:
         print("\nPlease correct the errors before proceeding!")
         sys.exit(1)
@@ -321,12 +372,11 @@ def update(force):
 def test():
     """ Tests some common pitfalls.
     """
-    model = get_model()
-    if tests(model, Hop.package_name):
+    if tests(HOP.model, HOP.package_name):
         click.echo('Tests OK')
     else:
         click.echo('Tests failed')
 
 
 if __name__ == '__main__':
-    main(obj={})
+    main({}, None)
